.btn1 {
  font-family: 'Arial', Times, serif;
  margin-right: 50px;
  width: 300px;
  height: 80px;
  font-size: 30px;
  font-weight: 1000;
  background-color: rgb(255, 126, 126);
  /* color: rgb(0, 0, 126); */
  border-radius: 25px;
  border: 5px solid rgb(255, 0, 173);
  /* box-shadow: 3px 3px rgb(98, 151, 98); */
}
